
import pytest
import numpy as np

from h5cross.same_nob import h5_same
from h5cross.same_nob import _same_nob


def test_same_nob():
    """ unit tests for base method"""
    ok_, log_o = _same_nob({"float": 1.5}, {"float": 1.5})
    dif, log_d = _same_nob({"float": 1}, {"float": 1.5})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"int": 1}, {"int": 1})
    dif, log_d = _same_nob({"int": 3}, {"int": 1})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"string": "string"}, {"string": "string"})
    dif, log_d = _same_nob({"string": "str"}, {"string": "string"})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"bytes": b"bytes"}, {"bytes": b"bytes"})
    dif, log_d = _same_nob({"bytes": "str"}, {"bytes": b"bytes"})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"bool": False}, {"bool": False})
    dif, log_d = _same_nob({"bool": True}, {"bool": False})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"complex": 1 + 5j}, {"complex": 1 + 5j})
    dif, log_d = _same_nob({"complex": 1}, {"complex": 1 + 5j})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"tuple": (1, 2, 3)}, {"tuple": (1, 2, 3)})
    dif, log_d = _same_nob({"tuple": (1, 5)}, {"tuple": (1, 2, 3)})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"list": [1, 2, 3]}, {"list": [1, 2, 3]})
    dif, log_d = _same_nob({"list": [0, 1]}, {"list": [1, 2, 3]})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"numpy": {"int32": np.int32(1)}}, {"numpy": {"int32": np.int32(1)}})
    dif, log_d = _same_nob({"numpy": {"int32": np.int32(3)}}, {"numpy": {"int32": np.int32(1)}})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"uint64": np.uint64(1)}, {"uint64": np.uint64(1)})
    dif, log_d = _same_nob({"uint64": np.uint64(3)}, {"uint64": np.uint64(1)})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"float": np.float_(1.5)}, {"float": np.float_(1.5)})
    dif, log_d = _same_nob({"float": np.float_(5)}, {"float": np.float_(1.5)})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"string": np.string_("string")}, {"string": np.string_("string")})
    dif, log_d = _same_nob({"string": np.string_("bool")}, {"string": np.string_("string")})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"bool": np.bool_(False)}, {"bool": np.bool_(False)})
    dif, log_d = _same_nob({"bool": np.bool_(True)}, {"bool": np.bool_(False)})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"bytes": np.bytes_("bytes")}, {"bytes": np.bytes_("bytes")})
    dif, log_d = _same_nob({"bytes": np.bytes_("bit")}, {"bytes": np.bytes_("bytes")})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"complex": np.complex_(1 + 5j)}, {"complex": np.complex_(1 + 5j)})
    dif, log_d = _same_nob({"complex": np.complex_(1 + 7j)}, {"complex": np.complex_(1 + 5j)})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"list": [np.int32(1), np.int32(2), np.int32(3)]}, {"list": [np.int32(1), np.int32(2), np.int32(3)]})
    dif, log_d = _same_nob({"list": [np.int32(1), np.uint64(2)]}, {"list": [np.int32(1), np.int32(2), np.int32(3)]})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"tuple": (np.int32(1), np.int32(2), np.int32(3))}, {"tuple": (np.int32(1), np.int32(2), np.int32(3))})
    dif, log_d = _same_nob({"tuple": (np.int32(1), np.int32(2), 5)}, {"tuple": (np.int32(1), np.int32(2), np.int32(3))})
    assert ok_, log_o
    assert not dif, log_d

    ok_, log_o = _same_nob({"ndarray": np.array([1, 2, 3])}, {"ndarray": np.array([1, 2, 3])})
    dif, log_d = _same_nob({"ndarray": np.array([1, 5, 3])}, {"ndarray": np.array([1, 2, 3])})
    assert ok_, log_o
    assert not dif, log_d

    dif, log_d = _same_nob({"ndarray": np.array([1, 3])}, {"ndarray": np.array([1, 2, 3])})
    assert not dif, log_d

    ok_, log_o = _same_nob({"2darray": np.array([[1, 2, 3], [1, 2, 3]])}, {"2darray": np.array([[1, 2, 3], [1, 2, 3]])})
    dif, log_d = _same_nob({"2darray": np.array([[1, 5, 3], [1, 5, 3]])}, {"2darray": np.array([[1, 2, 3], [1, 2, 3]])})
    assert ok_, log_o
    assert not dif, log_d

    class NotImpl():
        def __init__(self):
            self._not_understood = None

    NI = NotImpl()

    with pytest.raises(NotImplementedError):
        assert _same_nob({"notimpl": NI}, {"notimpl": NI})
        assert _same_nob({"notinlist": [NI, NI]}, {"notinlist": [NI, NI]})


def test_h5same(datadir):
    """test for h5 to h5 comparisons"""
    assert h5_same(datadir.join("source.mesh.h5"), datadir.join("source.mesh.h5"))
    with pytest.raises(ValueError):
        assert h5_same(datadir.join("source.mesh.h5"), datadir.join('target.mesh.h5'))





from h5cross.utils import _group_keys



# TODO: complete tests


def test_group_keys():
    keys = ['a/b/c', 'a/b/d', 'a/e', 'f/g', 'h']

    grouped_keys = _group_keys(keys, sep='/')

    target_results = {'a': {'b': {'c': 'a/b/c', 'd': 'a/b/d'}, 'e': 'a/e'},
                      'f': {'g': 'f/g'}, 'h': 'h'}
    assert grouped_keys['a']['b']['c'] == target_results['a']['b']['c']
    assert grouped_keys['a']['b']['d'] == target_results['a']['b']['d']
    assert grouped_keys['f']['g'] == target_results['f']['g']
    assert grouped_keys['h'] == target_results['h']

    # ensure order (root)
    target_keys = ['a', 'f', 'h']
    for key, target_key in zip(grouped_keys, target_keys):
        assert key == target_key

    # ensure order (nested)
    target_keys = ['b', 'e']
    for key, target_key in zip(grouped_keys['a'], target_keys):
        assert key == target_key




import os

import pytest

from h5cross.visit_h5 import get_h5_structure
from h5cross.visit_h5 import _ascii2string
from h5cross.io import write_h5
from h5cross.utils import unwrap_dict

from .data import TEST_DICT


def _create_example_h5(tmpdir, as_dict=False):
    filename = os.path.join(tmpdir, 'example')
    write_h5(TEST_DICT, filename)
    return get_h5_structure(f'{filename}.h5', as_dict=as_dict)


def test_ascii2string():
    word = "testing/"
    ascii_list = []
    for letter in word:
        ascii_list.append(ord(letter))

    assert word[:-1] == _ascii2string(ascii_list)


def test_get_structure(tmpdir):

    h5_struct_unwrapped = unwrap_dict(_create_example_h5(tmpdir, as_dict=False))

    # check keys
    target_struct_keys = ['Extra/H2 mass fraction', 'Extra/Extra/O2 mass fraction', 'pressure']

    for key in target_struct_keys:
        assert key in h5_struct_unwrapped


@pytest.mark.parametrize('as_dict', [False, True])
def test_get_structure_field(tmpdir, as_dict):
    # TODO: test non float
    h5_struct = _create_example_h5(tmpdir, as_dict=as_dict)

    # on root
    pressure_field = h5_struct['pressure']
    target_value = {'dtype': 'float64', 'value': 'array of 16 elements'}
    for key, target_value_ in target_value.items():
        if as_dict:
            assert pressure_field[key] == target_value_
        else:
            assert getattr(pressure_field, key) == target_value_

    # test one nested
    h2_field = h5_struct['Extra']['H2 mass fraction']
    target_value = {'dtype': 'float64', 'value': 'array of 20 elements'}
    for key, target_value_ in target_value.items():
        if as_dict:
            assert h2_field[key] == target_value_
        else:
            assert getattr(h2_field, key) == target_value_



import os
import datetime


import pytest
import numpy as np
import h5py

from h5cross import hdfdict


def equaldt(a, b):
    d = a - b
    return d.total_seconds() < 1e-3


@pytest.mark.parametrize("lazy", [True, False])
def test_dict_to_hdf(tmpdir, lazy):

    test_dict = {
        "a": np.random.randn(10),
        "b": [1, 2, 3],
        "c": "Hallo",  # TODO: check how to go to string automatically
        "d": np.array(["a", "b"]).astype("S"),
        "e": True,
        "f": (True, False),
        "g": [datetime.datetime.now() for i in range(5)],
        "h": datetime.datetime.utcnow(),
        # "i": [("Hello", 5), (6, "No HDF but json"), {"foo": True}],
        # "j": {"test2": datetime.datetime.now(), "again": ["a", 1], (1, 2): (3, 4)},
    }

    # dump
    filename = os.path.join(tmpdir, "example.h5")
    with h5py.File(filename, "w") as h5file:
        hdfdict.dump(test_dict, h5file)

    # load
    h5file = h5py.File(filename, "r")
    res = hdfdict.load(h5file, lazy=lazy)

    assert tuple(test_dict.keys()) == tuple(res.keys())
    np.testing.assert_allclose(test_dict["a"], res["a"])
    np.testing.assert_array_equal(test_dict["b"], res["b"])
    assert test_dict["c"] == res["c"].decode("utf-8")
    np.testing.assert_array_equal(test_dict["d"], res["d"])
    np.testing.assert_array_equal(test_dict["e"], res["e"])
    np.testing.assert_array_equal(test_dict["f"], res["f"])

    for a, b in zip(test_dict["g"], res["g"]):
        assert equaldt(a, b)

    assert equaldt(test_dict["h"], res["h"])

    # assert test_dict["i"][0][0] == "Hello"
    # assert test_dict["i"][1][0] == 6

    # assert isinstance(test_dict["j"]["test2"], datetime.datetime)
    # assert test_dict["j"]["again"][0] == "a"
    # assert res["j"]["again"][1] == 1

    if lazy:
        res.unlazy()  # all lazy objects will be rolled out.

    h5file.close()




import pytest

import numpy as np

from h5cross.stats import compute_dict_stats
from h5cross.prettytable import get_stats_table

from .data import TEST_DICT


ATOL = 1e-6


def test_stats():

    skip_list = ['pressure', 'O2 mass fraction']
    stats = compute_dict_stats(TEST_DICT, skip_list=skip_list)

    target_stats = {'min': 0.08493008, 'max': 0.93432919, 'mean': 0.532405342,
                    'median': 0.57242786, 'std': 0.2690518441418391}

    # ensure statistics are well computed
    h2_stats = stats['Extra']['H2 mass fraction']
    for key, target_value in target_stats.items():
        assert abs(target_value - getattr(h2_stats, key)) < ATOL

    # ensure skip list works
    assert 'pressure' not in stats
    assert stats['Extra'].get('Extra', None) is None


@pytest.mark.parametrize('full_path', [True, False])
def test_stats_table(full_path):

    stats_dict = compute_dict_stats(TEST_DICT)

    stats_table = get_stats_table(stats_dict, full_path=full_path)

    rows = stats_table._rows
    row_names = [row[0] for row in rows]
    target_row_names = ['Extra/H2 mass fraction', 'Extra/Extra/O2 mass fraction',
                        'pressure']

    assert len(rows) == 3
    assert len(rows[0]) == 6
    for row_name, target_row_name in zip(row_names, target_row_names):
        if full_path:
            assert row_name == target_row_name
        else:
            assert row_name == target_row_name.split('/')[-1]


def test_stats_array_type():

    array = np.array(['1', '2'])
    out = compute_dict_stats({'string': array})

    assert out == {}





import os

import yaml


from h5cross.io import write_dict_to_yaml
from h5cross.io import write_h5
from h5cross.io import load_hdf_as_dict


TEST_DICT = {'a': 1,
             'b': {'c': 1.2,
                   'd': 'a string'},
             'e': None}


def _assert_dict(result_dict):
    assert result_dict['a'] == TEST_DICT['a']
    assert result_dict['b']['c'] == TEST_DICT['b']['c']

    val = result_dict['b']['d']
    if type(val) is not str:
        assert val.decode('utf-8') == TEST_DICT['b']['d']
    else:
        assert val == TEST_DICT['b']['d']
    assert result_dict['e'] == TEST_DICT['e']


def test_write_load_dict_yaml(tmpdir):

    filename = os.path.join(tmpdir, 'example')

    write_dict_to_yaml(TEST_DICT, filename)

    # load file back
    with open(f'{filename}.yml', 'r', encoding='utf-8') as file:
        data_dict = yaml.load(file, Loader=yaml.SafeLoader)

    _assert_dict(data_dict)


def test_write_load_hdf(tmpdir):

    filename = os.path.join(tmpdir, 'example')
    write_h5(TEST_DICT, filename)

    data_dict = load_hdf_as_dict(f'{filename}.h5')
    _assert_dict(data_dict)




import os

from click.testing import CliRunner
import pytest


from h5cross import cli


@pytest.mark.parametrize('filenames', [['file1.h5', 'file2.h5'],
                                       ['file1.h5', 'file3.h5']])
@pytest.mark.parametrize('add_stats', [[], ['--add-stats']])
def test_compare_h5(datadir, filenames, add_stats):
    # just checks cli does not fail
    filenames = [os.path.join(datadir, filename) for filename in filenames]

    runner = CliRunner()
    result = runner.invoke(cli.diff, [*filenames, *add_stats, '--test'])

    assert result.exit_code == 0


def test_visual_h5(datadir):
    filename = os.path.join(datadir, 'file1.h5')

    runner = CliRunner()
    result = runner.invoke(cli.view, [filename, '--test'])

    assert result.exit_code == 0


@pytest.mark.parametrize('filenames', [['file1.h5', 'file2.h5'],
                                       ['file1.h5', 'file3.h5']])
@pytest.mark.parametrize('select_vars', [['--select-vars', 'temperature'],
                                         ['--select-vars', 'temperature,pressure']])
@pytest.mark.parametrize('use_seaborn', [['--use-seaborn'], []])
@pytest.mark.parametrize('add_xy', [['--add-xy'], []])
def test_compare_scatter_h5(datadir, tmpdir, filenames, select_vars, use_seaborn,
                            add_xy):
    filename = os.path.join(tmpdir, 'plot')

    filenames = [os.path.join(datadir, filename) for filename in filenames]

    runner = CliRunner()
    result = runner.invoke(cli.scatter,
                           [*filenames, *select_vars, *use_seaborn,
                            *add_xy, '--test', '--save-name', filename])

    assert result.exit_code == 0

    file_exists = False
    for cmp_filename in os.listdir(tmpdir):
        if cmp_filename.startswith('plot'):
            file_exists = True
            break

    assert file_exists is True




import os
import pytest
from h5cross.convert import convert


@pytest.mark.skip("External file dependent test. To change...")
def test_nek5000(meshes_dir):
    filename = os.path.join(meshes_dir, 'nek5000', 'channel3D_0.f00001')

    data_dict = convert(filename, file_format='nek5000')

    expected_keys = {'xmesh', 'ymesh', 'zmesh', 'ux', 'uy', 'uz', 'pressure',
                     'Parameters'}
    for key in data_dict:
        assert key in expected_keys


@pytest.mark.skip("External file dependent test. To change...")
def test_vtk(meshes_dir):
    filename = os.path.join(meshes_dir, 'vtk', 'triangle_mesh_linear.vtk')

    data_dict = convert(filename, file_format='vtk')

    expected_keys = {'pressure', 'velocity', 'Mesh'}
    for key in data_dict:
        assert key in expected_keys


@pytest.mark.skip("External file dependent test. To change...")
def test_vtu(meshes_dir):
    filename = os.path.join(meshes_dir, 'vtu', 'triangle_mesh_linear.vtu')

    data_dict = convert(filename, file_format='vtu')

    expected_keys = {'pressure', 'velocity', 'Mesh'}
    for key in data_dict:
        assert key in expected_keys


@pytest.mark.skip("External file dependent test. To change...")
def test_pvtu(meshes_dir):
    filename = os.path.join(meshes_dir, 'pvtu', 'example_pvtu.pvtu')

    data_dict = convert(filename, file_format='pvtu')

    expected_keys = {'node_value', 'simerr_type', 'Mesh'}
    for key in data_dict:
        assert key in expected_keys



"""Helps h5py to dump and load dictionaries.

Adapted from https://github.com/SiggiGue/hdfdict due to inactivity of the
developer.
"""

from pathlib import Path
from collections import UserDict
from datetime import datetime
from contextlib import contextmanager

import h5py
import yaml

try:
    from numpy import string_
except ImportError:
    from numpy import bytes_ as string_


TYPEID = "_type_"


@contextmanager
def _hdf_file(hdf, lazy=True, *args, **kwargs):
    """Context manager yields h5 file if hdf is str,
    otherwise just yield hdf as is."""
    if isinstance(hdf, str):
        if not lazy:
            with h5py.File(hdf, *args, **kwargs) as hdf:
                yield hdf
        else:
            yield h5py.File(hdf, *args, **kwargs)
    else:
        yield hdf


def _unpack_dataset(item):
    """Reconstruct a hdfdict dataset.
    Only some special unpacking for yaml and datetime types.

    Parameters
    ----------
    item : h5py.Dataset

    Returns
    -------
    value : Unpacked Data

    """
    value = item[()]

    if TYPEID in item.attrs:
        if item.attrs[TYPEID].astype(str) == "datetime":
            if hasattr(value, "__iter__"):
                value = [datetime.fromtimestamp(ts) for ts in value]
            else:
                value = datetime.fromtimestamp(value)

        if item.attrs[TYPEID].astype(str) == "yaml":
            value = yaml.safe_load(value.decode())

    return value


def _pack_dataset(hdfobject, key, value):
    """Packs a given key value pair into a dataset in the given hdfobject."""
    isdt = None
    if isinstance(value, datetime):
        value = value.timestamp()
        isdt = True

    if hasattr(value, "__iter__"):
        if all(isinstance(i, datetime) for i in value):
            value = [item.timestamp() for item in value]
            isdt = True

    try:
        ds = hdfobject.create_dataset(name=key, data=value)
        if isdt:
            ds.attrs.create(name=TYPEID, data=string_("datetime"))
    except TypeError:
        # Obviously the data was not serializable. To give it
        # a last try; serialize it to yaml
        # and save it to the hdf file:
        ds = hdfobject.create_dataset(name=key, data=string_(yaml.safe_dump(value)))
        ds.attrs.create(name=TYPEID, data=string_("yaml"))
        # if this fails again, restructure your data!


class LazyHdfDict(UserDict):
    """Helps loading data only if values from the dict are requested.

    This is done by reimplementing the __getitem__ method.
    """

    def __init__(self, _h5file=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._h5file = _h5file  # used to close the file on deletion.

    def __getitem__(self, key):
        """Returns item and loads dataset if needed."""
        item = super().__getitem__(key)
        if isinstance(item, h5py.Dataset):
            item = _unpack_dataset(item)
            self.__setitem__(key, item)
        return item

    def unlazy(self):
        """Unpacks all datasets.
        You can call dict(this_instance) then to get a real dict.
        """
        load(self, lazy=False)

    def close(self):
        """Closes the h5file if provided at initialization."""
        if self._h5file and hasattr(self._h5file, "close"):
            self._h5file.close()

    def __del__(self):
        self.close()

    def _ipython_key_completions_(self):
        """Returns a tuple of keys.
        Special Method for ipython to get key completion
        """
        return tuple(self.keys())


def load(hdf, lazy=True, unpacker=_unpack_dataset, *args, **kwargs):
    """Returns a dictionary containing the
    groups as keys and the datasets as values
    from given hdf file.

    Parameters
    ----------
    hdf : string (path to file) or `h5py.File()` or `h5py.Group()`
    lazy : bool
        If True, the datasets are lazy loaded at the moment an item is requested.
    upacker : callable
        Unpack function gets `value` of type h5py.Dataset.
        Must return the data you would like to have it in the returned dict.

    Returns
    -------
    d : dict
        The dictionary containing all groupnames as keys and
        datasets as values.
    """

    def _recurse(hdfobject, datadict):
        for key, value in hdfobject.items():
            if type(value) == h5py.Group or isinstance(value, LazyHdfDict):
                if lazy:
                    datadict[key] = LazyHdfDict()
                else:
                    datadict[key] = {}
                datadict[key] = _recurse(value, datadict[key])
            elif isinstance(value, h5py.Dataset):
                if not lazy:
                    value = unpacker(value)
                datadict[key] = value

        return datadict

    with _hdf_file(hdf, lazy=lazy, *args, **kwargs) as hdf:
        if lazy:
            data = LazyHdfDict(_h5file=hdf)
        else:
            data = {}
        return _recurse(hdf, data)


def dump(data, hdf, packer=_pack_dataset, *args, **kwargs):
    """Adds keys of given dict as groups and values as datasets
    to the given hdf-file (by string or object) or group object.

    Parameters
    ----------
    data : dict
        The dictionary containing only string keys and
        data values or dicts again.
    hdf : string (path to file) or `h5py.File()` or `h5py.Group()`
    packer : callable
        Callable gets `hdfobject, key, value` as input.
        `hdfobject` is considered to be either a h5py.File or a h5py.Group.
        `key` is the name of the dataset.
        `value` is the dataset to be packed and accepted by h5py.

    Returns
    -------
    hdf : obj
        `h5py.Group()` or `h5py.File()` instance

    """

    def _recurse(datadict, hdfobject):
        for key, value in datadict.items():
            if isinstance(value, (dict, LazyHdfDict)):
                hdfgroup = hdfobject.create_group(key)
                _recurse(value, hdfgroup)
            else:
                packer(hdfobject, key, value)

    with _hdf_file(hdf, *args, **kwargs) as hdf:
        _recurse(data, hdf)
        return hdf


# TYPE = "_type_"


# @contextmanager
# def hdf_file(hdf, lazy=True, *args, **kwargs):
#     """Context manager that yields an h5 file if `hdf` is a string,
#     otherwise it yields hdf as is."""
#     if isinstance(hdf, (str, Path)):
#         if not lazy:
#             # The file can be closed after reading
#             # therefore the context manager is used.
#             with h5py.File(hdf, *args, **kwargs) as hdf:
#                 yield hdf
#         else:
#             # The file should stay open because datasets
#             # are read on item access.
#             yield h5py.File(hdf, *args, **kwargs)
#     else:
#         yield hdf


# def unpack_dataset(item):
#     """Reconstruct a hdfdict dataset.
#     Only some special unpacking for yaml and datetime types.

#     Parameters
#     ----------
#     item : h5py.Dataset

#     Returns
#     -------
#     key: Unpacked key
#     value : Unpacked Data

#     """
#     value = item[()]
#     type_id = item.attrs.get(TYPE, string_()).astype(str)
#     if type_id == "datetime":
#         if hasattr(value, "__iter__"):
#             value = [datetime.fromtimestamp(ts) for ts in value]
#         else:
#             value = datetime.fromtimestamp(value)

#     elif type_id == "yaml":
#         value = yaml.safe_load(value.decode())

#     elif type_id == "list":
#         value = list(value)

#     elif type_id == "tuple":
#         value = tuple(value)

#     elif type_id == "str":
#         value = string_(value).astype(str)

#     return value


# class LazyHdfDict(UserDict):
#     """Helps loading data only if values from the dict are requested.

#     This is done by reimplementing the __getitem__ method.

#     """

#     def __init__(self, _h5file=None, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self._h5file = _h5file  # used to close the file on deletion.

#     def __getitem__(self, key):
#         """Returns item and loads dataset if needed."""
#         item = super().__getitem__(key)
#         if isinstance(item, h5py.Dataset):
#             item = unpack_dataset(item)
#             self.__setitem__(key, item)
#         return item

#     def unlazy(self):
#         """Unpacks all datasets.
#         You can call dict(this_instance) then to get a real dict.
#         """
#         load(self, lazy=False)

#     def close(self):
#         """Closes the h5file if provided at initialization."""
#         if self._h5file and hasattr(self._h5file, "close"):
#             self._h5file.close()

#     def __del__(self):
#         self.close()

#     def _ipython_key_completions_(self):
#         """Returns a tuple of keys.
#         Special Method for ipython to get key completion
#         """
#         return tuple(self.keys())


# def load(hdf, lazy=True, unpacker=unpack_dataset, mode="r", *args, **kwargs):
#     """Returns a dictionary containing the
#     groups as keys and the datasets as values
#     from given hdf file.

#     Parameters
#     ----------
#     hdf : string (path to file) or `h5py.File()` or `h5py.Group()`
#     lazy : bool
#         If True, the datasets are lazy loaded at the moment an item is requested.
#     upacker : callable
#         Unpack function gets `value` of type h5py.Dataset.
#         Must return the data you would like to have it in the returned dict.
#     mode : str
#         File read mode. Default: 'r'.

#     Returns
#     -------
#     d : dict
#         The dictionary containing all groupnames as keys and
#         datasets as values.
#     """

#     def _recurse(hdfobject, datadict):
#         for key, value in hdfobject.items():
#             if type(value) == h5py.Group or isinstance(value, LazyHdfDict):
#                 if lazy:
#                     datadict[key] = LazyHdfDict()
#                 else:
#                     datadict[key] = {}
#                 datadict[key] = _recurse(value, datadict[key])
#             elif isinstance(value, h5py.Dataset):
#                 if not lazy:
#                     value = unpacker(value)
#                 datadict[key] = value

#         return datadict

#     with hdf_file(hdf, lazy=lazy, mode=mode, *args, **kwargs) as hdf:
#         if lazy:
#             data = LazyHdfDict(_h5file=hdf)
#         else:
#             data = {}
#         return _recurse(hdf, data)


# def pack_dataset(hdfobject, key, value):
#     """Packs a given key value pair into a dataset in the given hdfobject."""

#     isdt = None
#     if isinstance(value, datetime):
#         value = value.timestamp()
#         isdt = True

#     if hasattr(value, "__iter__"):
#         if all(isinstance(i, datetime) for i in value):
#             value = [item.timestamp() for item in value]
#             isdt = True

#     try:
#         ds = hdfobject.create_dataset(name=key, data=value)

#         if isdt:
#             attr_data = "datetime"
#         elif isinstance(value, list):
#             attr_data = "list"
#         elif isinstance(value, tuple):
#             attr_data = "tuple"
#         elif isinstance(value, str):
#             attr_data = "str"
#         else:
#             attr_data = None

#         if attr_data:
#             ds.attrs.create(name=TYPE, data=string_(attr_data))

#     except (TypeError, ValueError):
#         # Obviously the data was not serializable. To give it
#         # a last try; serialize it to yaml
#         # and save it to the hdf file:
#         ds = hdfobject.create_dataset(name=key, data=string_(yaml.safe_dump(value)))

#         ds.attrs.create(name=TYPE, data=string_("yaml"))
#         # if this fails again, restructure your data!


# def dump(data, hdf, packer=pack_dataset, mode="w", *args, **kwargs):
#     """Adds keys of given dict as groups and values as datasets
#     to the given hdf-file (by string or object) or group object.

#     Parameters
#     ----------
#     data : dict
#         The dictionary containing only string keys and
#         data values or dicts again.
#     hdf : string (path to file) or `h5py.File()` or `h5py.Group()`
#     packer : callable
#         Callable gets `hdfobject, key, value` as input.
#         `hdfobject` is considered to be either a h5py.File or a h5py.Group.
#         `key` is the name of the dataset.
#         `value` is the dataset to be packed and accepted by h5py.
#     mode : str
#         File write mode. Default: 'w'

#     Returns
#     -------
#     hdf : obj
#         `h5py.Group()` or `h5py.File()` instance

#     """

#     def _recurse(datadict, hdfobject):
#         for key, value in datadict.items():
#             if isinstance(value, (dict, LazyHdfDict)):
#                 hdfgroup = hdfobject.create_group(key)
#                 _recurse(value, hdfgroup)
#             else:
#                 packer(hdfobject, key, value)

#     with hdf_file(hdf, mode=mode, *args, **kwargs) as hdf:
#         _recurse(data, hdf)
#         return hdf
